

class DefaultSMSMessage(object):

    def __init__(self, notification):
        self.to = notification.target_user
        self.subject = self.kwargs.get('subject')
        self.body = ''

    def create_message(self):
        """
        """
        self.body = ""

    def send(self):
        pass

    def check_validity(self):
        return True



import logging

from django.core.cache import cache
from django.test import TestCase
from transmissions.lock import lock

class LockTests(TestCase):

    def setUp(self):
        logging.disable(logging.WARNING)

    def test_simple(self):

        v = 1
        with lock('simple'):
            v = 2

        self.assertEqual(v, 2)

    def test_fail(self):

        v = 1
        key = 'fail'
        lock_id = 'lock-transmission-{0}'.format(key)
        cache.add(lock_id, 1, 200)
        with self.assertRaisesMessage(RuntimeError, 'Lock could not be acquired after 110ms'):
            with lock(key, 100):
                v = 2
        cache.delete(lock_id)

        self.assertEqual(v, 1)




from django.core.cache import cache
import contextlib
import time


@contextlib.contextmanager
def lock(key, timeout=5000):
    """
    A simple context manager that raises the passed exception
    if a lock can't be acquired.
    """

    lock_id = 'lock-transmission-{0}'.format(key)
    acquire_lock = lambda: cache.add(lock_id, 1, timeout)
    release_lock = lambda: cache.delete(lock_id)

    waited, hops = 0, 10
    while not acquire_lock():
        time.sleep(float(hops) / 1000.0)
        waited += hops
        if waited > timeout:
            raise RuntimeError('Lock could not be acquired after {}ms'.format(waited))

    try:
        yield
    finally:
        release_lock()



from optparse import make_option
from django.utils import timezone
from django.core.management.base import BaseCommand
from transmissions.models import Notification
from transmissions.lock import lock


class Command(BaseCommand):
    help = 'Retry to process failed notifications'

    option_list = BaseCommand.option_list + (
        make_option('--days', action='store', dest='days', type='int', default=7,
            help='Number of days since last failure'),
    )

    def handle(self, days, db_dry_run=False, *args, **options):

        since_datetime = timezone.now() - timezone.timedelta(days=days)

        notifications = Notification.objects.filter(datetime_scheduled__lte=timezone.now(),
                                                    datetime_scheduled__gte=since_datetime,
                                                    status=Notification.Status.FAILED).order_by('datetime_scheduled')

        for notification in notifications:
            self.stdout.write("Sending {}".format(notification))
            with lock('{0}'.format(notification.id)):
                notification.send()



import logging

from django.utils import timezone

from transmissions.exceptions import DuplicateNotification
from transmissions.lock import lock

register = {}


def message(trigger_name, behavior=None, **kwargs):
    def wrapper(cls):
        from django.contrib.contenttypes.models import ContentType
        from transmissions.models import TriggerBehavior, Notification

        cls.trigger_name = trigger_name
        if behavior in TriggerBehavior.values.keys():
            cls.behavior = behavior
        else:
            cls.behavior = TriggerBehavior.DEFAULT
        cls.kwargs = kwargs

        if trigger_name in register:
            logging.getLogger('django-transmissions').warning(
                'Duplicate definition for trigger {} at {} and {}.{}',
                trigger_name, register[trigger_name],
                cls.__module__, cls.__name__)

        register[trigger_name] = "{}.{}".format(cls.__module__, cls.__name__)

        def trigger(cls, target_user, trigger_user=None,
                    datetime_scheduled=None, content=None, data=None, silent=True):
            """
            Trigger a notification
            """

            # No need for a lock
            if cls.behavior in (TriggerBehavior.DEFAULT, TriggerBehavior.DELETE_AFTER_PROCESSING):
                return _trigger_within_lock(cls,
                                            target_user,
                                            trigger_user,
                                            datetime_scheduled,
                                            content,
                                            data,
                                            silent)

            # Acquire lock before triggering
            else:
                key = '{}@{}'.format(cls.trigger_name, target_user.id)
                if cls.behavior in (TriggerBehavior.SEND_ONCE_PER_CONTENT,
                                    TriggerBehavior.TRIGGER_ONCE_PER_CONTENT):
                    key += '+{}.{}'.format(content.__class__.__name__, content.id)

                with lock(key):
                    return _trigger_within_lock(cls,
                                                target_user,
                                                trigger_user,
                                                datetime_scheduled,
                                                content,
                                                data,
                                                silent)

        def _trigger_within_lock(cls, target_user, trigger_user=None,
                                 datetime_scheduled=None, content=None, data=None, silent=True):

            try:
                if (cls.behavior == TriggerBehavior.SEND_ONCE and
                        Notification.objects.filter(
                            target_user=target_user,
                            trigger_name=cls.trigger_name).exists()):
                    raise DuplicateNotification()

                if (cls.behavior == TriggerBehavior.SEND_ONCE_PER_CONTENT and
                        Notification.objects.filter(
                            target_user=target_user,
                            trigger_name=cls.trigger_name,
                            content_type=ContentType.objects.get_for_model(content),
                            content_id=content.id).exists()):
                    raise DuplicateNotification()

                if (cls.behavior == TriggerBehavior.TRIGGER_ONCE and
                        Notification.objects.filter(target_user=target_user,
                                                    trigger_name=cls.trigger_name,
                                                    datetime_processed__isnull=True).exists()):
                    raise DuplicateNotification()

                if (cls.behavior == TriggerBehavior.TRIGGER_ONCE_PER_CONTENT and
                        Notification.objects.filter(
                            target_user=target_user,
                            trigger_name=cls.trigger_name,
                            datetime_processed__isnull=True,
                            content_type=ContentType.objects.get_for_model(content),
                            content_id=content.id).exists()):
                    raise DuplicateNotification()

                if cls.behavior == TriggerBehavior.LAST_ONLY:
                    waiting_notifications = Notification.objects.filter(
                        target_user=target_user,
                        trigger_name=cls.trigger_name,
                        datetime_processed__isnull=True)
                    for waiting_notification in waiting_notifications:
                        waiting_notification.cancel()

            except DuplicateNotification:
                if not silent:
                    raise
                else:
                    return None

            if datetime_scheduled is None:
                datetime_scheduled = timezone.now()

            extra = {}
            if content is not None:
                extra['content'] = content
            if data is not None:
                extra['data'] = data

            notification = Notification.objects.create(trigger_name=cls.trigger_name,
                                                       target_user=target_user,
                                                       trigger_user=trigger_user,
                                                       datetime_scheduled=datetime_scheduled,
                                                       status=Notification.Status.CREATED,
                                                       **extra)
            return notification

        cls.trigger = classmethod(trigger)

        return cls

    return wrapper



import logging
from datetime import date, time

import pytz
from django.test import TestCase
from django.utils import timezone
from django.core.management import call_command

import mock
from transmissions import message
from transmissions.models import Notification
from transmissions.channels.email import DefaultEmailMessage
from . import factories


@message('command_test', behavior=None, subject='Hello World!')
class FailedAttemptMessage(DefaultEmailMessage):
    template_name = 'test'

class CommandTests(TestCase):

    def setUp(self):
        logging.disable(logging.WARNING)

    def test_simple(self):

        trigger_name = 'command_test'
        user = factories.User()

        # Trigger notification
        notification = FailedAttemptMessage.trigger(user, datetime_scheduled=None)
        self.assertEqual(notification.trigger_name, trigger_name)
        self.assertEqual(notification.target_user, user)
        self.assertEqual(notification.status, Notification.Status.CREATED)

        # Test
        call_command('retry_failed_notifications', days=8)
        notification = Notification.objects.get(pk=notification.id)
        self.assertEqual(notification.status, Notification.Status.CREATED)

        # Update to failed
        notification.status = Notification.Status.FAILED
        notification.save()

        # Test
        call_command('retry_failed_notifications', days=8)
        notification = Notification.objects.get(pk=notification.id)
        self.assertEqual(notification.status, Notification.Status.SUCCESSFULLY_SENT)


    @mock.patch.object(timezone, 'now')
    def test_older(self, mock_timezone_now):

        # Mock time to now
        mock_time = time(8, 0, tzinfo=pytz.utc)
        today = date.today()
        mock_timezone_now.return_value = timezone.datetime.combine(today, mock_time)

        trigger_name = 'command_test'
        user = factories.User()

        # Trigger notification
        notification = FailedAttemptMessage.trigger(user, datetime_scheduled=None)
        self.assertEqual(notification.trigger_name, trigger_name)
        self.assertEqual(notification.target_user, user)
        self.assertEqual(notification.status, Notification.Status.CREATED)

        # Test
        call_command('retry_failed_notifications', days=8)
        notification = Notification.objects.get(pk=notification.id)
        self.assertEqual(notification.status, Notification.Status.CREATED)

        # Update to failed
        notification.status = Notification.Status.FAILED
        notification.save()

        # Move 10 days in the future
        mock_timezone_now.return_value += timezone.timedelta(days=10)

        # Test
        call_command('retry_failed_notifications', days=8)
        notification = Notification.objects.get(pk=notification.id)
        self.assertEqual(notification.status, Notification.Status.FAILED)


import logging

from django.test import TestCase
from django.utils import timezone

from transmissions.models import Notification, TriggerBehavior
from transmissions.channels.email import DefaultEmailMessage
from transmissions import exceptions, message
from . import factories


TRIGGER_SIMPLE = 'simple-model'
@message(TRIGGER_SIMPLE)
class SimpleMessage(DefaultEmailMessage):
    template_name = 'test'

TRIGGER_SEND_ONCE = 'send_once'
@message(TRIGGER_SEND_ONCE, behavior=TriggerBehavior.SEND_ONCE)
class SendOnceMessage(DefaultEmailMessage):
    template_name = 'test'

TRIGGER_SEND_ONCE_PER_CONTENT = 'send_once_per_content'
@message(TRIGGER_SEND_ONCE_PER_CONTENT, behavior=TriggerBehavior.SEND_ONCE_PER_CONTENT)
class SendOncePerContentMessage(DefaultEmailMessage):
    template_name = 'test'

TRIGGER_ONCE = 'trigger_once'
@message(TRIGGER_ONCE, behavior=TriggerBehavior.TRIGGER_ONCE)
class TriggerOnceMessage(DefaultEmailMessage):
    template_name = 'test'

TRIGGER_ONCE_PER_CONTENT = 'trigger_once_per_content'
@message(TRIGGER_ONCE_PER_CONTENT, behavior=TriggerBehavior.TRIGGER_ONCE_PER_CONTENT)
class TriggerOncePerContentMessage(DefaultEmailMessage):
    template_name = 'test'

TRIGGER_DELETE_AFTER_PROCESSING = 'trigger_delete_after'
@message(TRIGGER_DELETE_AFTER_PROCESSING, behavior=TriggerBehavior.DELETE_AFTER_PROCESSING)
class TriggerDeleteAfterMessage(DefaultEmailMessage):
    template_name = 'test'

class ModelTests(TestCase):

    def setUp(self):
        logging.disable(logging.WARNING)


    def test_trigger_notification(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            notification = SimpleMessage.trigger(user)

            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)
            self.assertIsNone(notification.content)

            diff = notification.datetime_created - notification.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)

    def test_trigger_notification_with_data(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            data = {'hello': 'World', 'count': 5, 'date': timezone.now().date()}
            notification = SimpleMessage.trigger(user, data=data)

            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)
            self.assertIsNone(notification.content)

            diff = notification.datetime_created - notification.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)

            notification = Notification.objects.get(pk=notification.id)
            self.assertEqual(notification.data, data)

    def test_trigger_notification_send_once(self):

        welcome_path = "{}.{}".format(SendOnceMessage.__module__, SendOnceMessage.__name__)
        trigger_settings = {TRIGGER_SEND_ONCE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            # First notification
            notification = SendOnceMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_SEND_ONCE)
            self.assertEqual(notification.target_user, user)

            # Second notification
            with self.assertRaises(exceptions.DuplicateNotification):
                SendOnceMessage.trigger(user, silent=False)

    def test_trigger_notification_send_once_per_content(self):

        welcome_path = "{}.{}".format(SendOncePerContentMessage.__module__, SendOncePerContentMessage.__name__)
        trigger_settings = {TRIGGER_SEND_ONCE_PER_CONTENT: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()
            other_user = factories.User()

            # First notification
            notification = SendOncePerContentMessage.trigger(user, content=user)
            self.assertEqual(notification.trigger_name, TRIGGER_SEND_ONCE_PER_CONTENT)
            self.assertEqual(notification.target_user, user)
            self.assertEqual(notification.content, user)

            # Second notification with different content
            notification = SendOncePerContentMessage.trigger(user, content=other_user)
            self.assertEqual(notification.trigger_name, TRIGGER_SEND_ONCE_PER_CONTENT)
            self.assertEqual(notification.target_user, user)
            self.assertEqual(notification.content, other_user)

            # Second notification with the same content
            with self.assertRaises(exceptions.DuplicateNotification):
                SendOncePerContentMessage.trigger(user, content=user, silent=False)

    def test_trigger_notification_trigger_once(self):

        welcome_path = "{}.{}".format(TriggerOnceMessage.__module__, TriggerOnceMessage.__name__)
        trigger_settings = {TRIGGER_ONCE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            # First notification
            notification = TriggerOnceMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_ONCE)
            self.assertEqual(notification.target_user, user)

            # Second notification while 1st is not processed yet
            with self.assertRaises(exceptions.DuplicateNotification):
                TriggerOnceMessage.trigger(user, silent=False)

            # Mark 1st notification as processed
            notification.datetime_processed = timezone.now()
            notification.save()

            # Second notification while 1st IS NOW processed
            new_notification = TriggerOnceMessage.trigger(user)
            self.assertNotEqual(new_notification, notification)
            self.assertEqual(new_notification.trigger_name, TRIGGER_ONCE)
            self.assertEqual(new_notification.target_user, user)

    def test_trigger_notification_trigger_once_per_content(self):

        welcome_path = "{}.{}".format(TriggerOncePerContentMessage.__module__, TriggerOncePerContentMessage.__name__)
        trigger_settings = {TRIGGER_ONCE_PER_CONTENT: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()
            other_user = factories.User()

            # First notification
            notification = TriggerOncePerContentMessage.trigger(user, content=user)
            self.assertEqual(notification.trigger_name, TRIGGER_ONCE_PER_CONTENT)
            self.assertEqual(notification.target_user, user)
            self.assertEqual(notification.content, user)

            # Second notification while 1st is not processed yet with different content
            other_notification = TriggerOncePerContentMessage.trigger(user, content=other_user)
            self.assertEqual(other_notification.trigger_name, TRIGGER_ONCE_PER_CONTENT)
            self.assertEqual(other_notification.target_user, user)
            self.assertEqual(other_notification.content, other_user)

            # Second notification while 1st is not processed yet
            with self.assertRaises(exceptions.DuplicateNotification):
                TriggerOncePerContentMessage.trigger(user, content=user, silent=False)

            # Mark 1st notification as processed
            notification.datetime_processed = timezone.now()
            notification.save()

            # Second notification while 1st IS NOW processed
            new_notification = TriggerOncePerContentMessage.trigger(user, content=user)
            self.assertNotEqual(new_notification, notification)
            self.assertEqual(new_notification.trigger_name, TRIGGER_ONCE_PER_CONTENT)
            self.assertEqual(new_notification.target_user, user)
            self.assertEqual(new_notification.content, user)

    def test_trigger_notification_delete_after_processing(self):

        welcome_path = "{}.{}".format(TriggerDeleteAfterMessage.__module__, TriggerDeleteAfterMessage.__name__)
        trigger_settings = {TRIGGER_DELETE_AFTER_PROCESSING: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            # First notification
            notification = TriggerDeleteAfterMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_DELETE_AFTER_PROCESSING)
            self.assertEqual(notification.target_user, user)

            # Second notification while 1st is not processed yet
            second_notification = TriggerDeleteAfterMessage.trigger(user)
            self.assertEqual(second_notification.trigger_name, TRIGGER_DELETE_AFTER_PROCESSING)
            self.assertEqual(second_notification.target_user, user)

            # Mark 1st notification as processed
            notification.send()

            with self.assertRaises(Notification.DoesNotExist):
                Notification.objects.get(pk=notification.id)

            second_notification = Notification.objects.get(pk=second_notification.id)
            self.assertEqual(second_notification.status, Notification.Status.CREATED)



# -*- coding: utf-8 -*-
"""
    django-transmissions.factories
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Factories used by unit tests and sample code

"""
from factory import DjangoModelFactory, Sequence
from django.conf import settings


class User(DjangoModelFactory):

    class Meta:
        model = settings.AUTH_USER_MODEL

    username = Sequence(lambda n: 'factory_user_{}'.format(n))
    email = Sequence(lambda n: 'factory-user-{}@example.com'.format(n))



import logging

from django.test import TestCase

from transmissions import message
from transmissions.models import Notification
from transmissions.channels.email import DefaultEmailMessage
from . import factories


TRIGGER_NAME = 'package_test'

@message(TRIGGER_NAME, behavior=None)
class PackageTestMessage(DefaultEmailMessage):
    template_name = 'test'


class PackageTests(TestCase):
    def setUp(self):
        logging.disable(logging.WARNING)

    def test_trigger(self):

        welcome_path = "{}.{}".format(PackageTestMessage.__module__, PackageTestMessage.__name__)
        trigger_settings = {TRIGGER_NAME: welcome_path}

        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            # Create user + trigger
            u = factories.User()

            # Trigger notification
            notification = PackageTestMessage.trigger(u)

            # Check notification
            self.assertEqual(notification.trigger_name, TRIGGER_NAME)
            self.assertEqual(notification.target_user, u)
            self.assertEqual(notification.status, Notification.Status.CREATED)


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('transmissions', '0002_auto_20150105_1523'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='datetime_processed',
            field=models.DateTimeField(null=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='notification',
            name='datetime_scheduled',
            field=models.DateTimeField(db_index=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    if hasattr(settings, 'TRANSMISSION_USER_MODEL'):
        USER_MODEL = settings.TRANSMISSION_USER_MODEL
    else:
        USER_MODEL = settings.AUTH_USER_MODEL

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', django_extensions.db.fields.ShortUUIDField(unique=True, editable=False, blank=True)),
                ('trigger_name', models.CharField(max_length=50, db_index=True)),
                ('content_id', models.PositiveIntegerField(null=True, blank=True)),
                ('data_pickled', models.TextField(editable=False, blank=True)),
                ('datetime_created', models.DateTimeField(auto_now_add=True, null=True)),
                ('datetime_scheduled', models.DateTimeField()),
                ('datetime_processed', models.DateTimeField(null=True)),
                ('datetime_seen', models.DateTimeField(null=True)),
                ('datetime_consumed', models.DateTimeField(null=True)),
                ('status', models.IntegerField(default=-2, db_index=True)),
                ('content_type', models.ForeignKey(blank=True, to='contenttypes.ContentType', null=True)),
                ('target_user', models.ForeignKey(related_name='notifications', to=USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Trigger',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=50)),
                ('message', models.TextField()),
                ('channel', models.IntegerField(default=1, db_index=True)),
                ('behavior', models.IntegerField(default=10, db_index=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='notification',
            name='trigger',
            field=models.ForeignKey(blank=True, to='transmissions.Trigger', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='notification',
            name='trigger_user',
            field=models.ForeignKey(related_name='notifications_sent', default=None, to=USER_MODEL, null=True),
            preserve_default=True,
        ),
        migrations.AlterIndexTogether(
            name='notification',
            index_together=set([('datetime_processed', 'datetime_scheduled'), ('target_user', 'datetime_scheduled'), ('target_user', 'trigger', 'datetime_processed')]),
        ),
    ]



import logging

from django.test import TestCase

from transmissions import message
from transmissions.channels import Channel
from transmissions.channels.email import DefaultEmailMessage
from . import factories


@message('welcome', behavior=None, subject='Hello World!')
class HelloWorld(DefaultEmailMessage):
    template_name = 'test'

class DecoratorTests(TestCase):

    def setUp(self):
        logging.disable(logging.WARNING)


    def test_simple(self):

        trigger_name = 'welcome'
        user = factories.User()

        welcome_path = "{}.{}".format(HelloWorld.__module__, HelloWorld.__name__)
        trigger_settings = {trigger_name: welcome_path}

        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):


            # Trigger notification
            notification = HelloWorld.trigger(user, datetime_scheduled=None)
            self.assertEqual(notification.trigger_name, trigger_name)
            self.assertEqual(notification.target_user, user)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, HelloWorld)


from django.contrib import admin
from transmissions.models import Notification


def link_to_field(field_name, short_description=None, getter=None):
    """
        Generic field linking for to-one.
    """

    def fn(obj):
        field = getter(obj) if getter else getattr(obj, field_name)

        if field:
            return '<a href="%s">%s</a>' % (field.get_admin_url(), field)
        return '(None)'

    fn.short_description = short_description or field_name.title()
    fn.admin_order_field = field_name
    fn.allow_tags = True
    return fn

_link_to_target_user = link_to_field('target_user')
_link_to_target_user.short_description = 'Target user'
_link_to_trigger_user = link_to_field('trigger_user')
_link_to_trigger_user.short_description = 'Trigger user'

class NotificationAdmin(admin.ModelAdmin):


    list_filter = ['trigger_name', 'status']

    date_hierarchy = 'datetime_processed'

    list_display = [
        'id',
        link_to_field('target_user'),
        'trigger_name',
        'status',
        'datetime_created',
        'datetime_scheduled',
        'datetime_processed'
    ]

    search_fields = ['=id', 'target_user__email']

    readonly_fields = ['id', 'trigger_name', _link_to_target_user, _link_to_trigger_user, 'datetime_created',
                       'datetime_processed', 'datetime_seen', 'datetime_consumed', 'content']

    fieldsets = (('Notifications', {'fields': ('id', _link_to_target_user, 'trigger_name')}),
                 ('Status', {'fields': (
                     'status', _link_to_trigger_user, 'datetime_created', 'datetime_scheduled', 'datetime_processed',
                     'datetime_seen',  'datetime_consumed')}),
                 ('Content', {'fields': ('content',)})
    )


admin.site.register(Notification, NotificationAdmin)


# -*- coding: utf-8 -*-
"""
    django-transmissions.channels.email_mandrill
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Mandrill email channel

"""
from django.core.mail import EmailMessage


class DefaultEmailMessage(object):

    def __init__(self, notification):
        self.to = notification.target_user
        self.subject = self.kwargs.get('subject')
        self.body = ''

    def create_message(self):
        """ Create EmailMessage object from message data

        :return: EmailMessage
        """
        msg = EmailMessage(self.subject, body=self.body, to=[self.to.email])
        return msg

    def send(self):
        self.create_message().send()

    def check_validity(self):
        """ Email is valid and can be sent """
        return True



# -*- coding: utf-8 -*-
"""
    django-transmissions.exceptions
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module contains all exceptions used by the Django Transmissions module
"""


class DuplicateNotification(Exception):
    """ Notification could not be triggered again because of trigger limitation """

class ChannelSendException(Exception):
    """ Notification could not be sent through this channel """

class UnknownTriggerException(Exception):
    """ Notification's trigger is not defined """


from django.core.management.base import BaseCommand
from django.db import connection
from django.db.utils import DatabaseError


class Command(BaseCommand):
    help = 'Rename app from django_transmissions to transmissions'
    option_list = BaseCommand.option_list
    args = ""

    def handle(self, db_dry_run=False, *args, **options):

        message = "Updating django_content_type app_label"
        sql = "UPDATE django_content_type SET app_label='transmissions' WHERE app_label='django_transmissions'"
        self.execute_sql(sql, message)

        message = "Rename django_transmissions_trigger"
        sql = "ALTER TABLE django_transmissions_trigger RENAME TO transmissions_trigger"
        self.execute_sql(sql, message)

        message = "Rename django_transmissions_notification"
        sql = "ALTER TABLE django_transmissions_notification RENAME TO transmissions_notification"
        self.execute_sql(sql, message)

    def execute_sql(self, sql, message):

        cursor = connection.cursor()

        self.stdout.write("Rename django_transmissions_notification... ", ending='')
        try:
            cursor.execute(sql)
            self.stdout.write('{}... SUCCEEDED'.format(message))
        except DatabaseError:
            self.stderr.write('{}... FAILED'.format(message))


import logging

from django.core import mail
from django.test import TestCase
from django.utils import timezone

from transmissions import tasks, message
from transmissions.channels import Channel
from transmissions.channels.email import DefaultEmailMessage
from transmissions.models import Notification
from . import factories

try:
    xrange
except NameError:
    xrange = range

TRIGGER_NAME = 'task_test'
TRIGGER_SUBJECT = 'This is a task test'

@message(TRIGGER_NAME, behavior=None, subject=TRIGGER_SUBJECT)
class TaskTestMessage(DefaultEmailMessage):
    template_name = 'test'


@message('broken', behavior=None, subject=TRIGGER_SUBJECT)
class BrokenMessage(DefaultEmailMessage):
    template_name = 'broken'

    def __init__(self, notification):
        super(BrokenMessage, self).__init__(notification)
        raise Exception('This message is broken at init')

class TasksTests(TestCase):

    def setUp(self):
        logging.disable(logging.WARNING)

        welcome_path = "{}.{}".format(TaskTestMessage.__module__, TaskTestMessage.__name__)
        self.trigger_settings = {TRIGGER_NAME: welcome_path}

    def test_process_task(self):

        with self.settings(TRANSMISSIONS_TRIGGERS=self.trigger_settings):
            user = factories.User()

            # Create trigger
            trigger_name = TRIGGER_NAME

            # Trigger notification
            notification = TaskTestMessage.trigger(user)
            self.assertEqual(notification.trigger_name, trigger_name)
            self.assertEqual(notification.target_user, user)
            self.assertEqual(notification.status, Notification.Status.CREATED)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, TaskTestMessage)

            # Trigger single task
            tasks.process_notification(notification.id)
            notification = Notification.objects.get(pk=notification.id)

            # Check status
            self.assertEqual(notification.status, Notification.Status.SUCCESSFULLY_SENT)
            self.assertGreaterEqual(notification.datetime_processed, notification.datetime_scheduled)

            # Check email was sent
            self.assertEqual(len(mail.outbox), 1)
            self.assertEqual(mail.outbox[0].subject, TRIGGER_SUBJECT)


    def test_process_task_multiple_times(self):

        with self.settings(TRANSMISSIONS_TRIGGERS=self.trigger_settings):
            user = factories.User()

            # Create trigger
            trigger_name = TRIGGER_NAME

            # Trigger notification
            notification = TaskTestMessage.trigger(user)
            self.assertEqual(notification.trigger_name, trigger_name)
            self.assertEqual(notification.target_user, user)
            self.assertEqual(notification.status, Notification.Status.CREATED)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, TaskTestMessage)

            # Trigger single task
            tasks.process_notification(notification.id)
            tasks.process_notification(notification.id)
            tasks.process_notification(notification.id)
            tasks.process_notification(notification.id)
            tasks.process_notification(notification.id)
            notification = Notification.objects.get(pk=notification.id)

            # Check status
            self.assertEqual(notification.status, Notification.Status.SUCCESSFULLY_SENT)
            self.assertGreaterEqual(notification.datetime_processed, notification.datetime_scheduled)

            # Check email was sent
            self.assertEqual(len(mail.outbox), 1)
            self.assertEqual(mail.outbox[0].subject, TRIGGER_SUBJECT)


    def test_process_all_notifications_task(self):

        with self.settings(TRANSMISSIONS_TRIGGERS=self.trigger_settings):
            users = [factories.User() for i in xrange(7)]
            notifications = []

            # Create trigger
            trigger_name = TRIGGER_NAME

            # Trigger notification
            for user in users:
                notification = TaskTestMessage.trigger(user)
                notifications.append(notification)
                self.assertEqual(notification.trigger_name, trigger_name)
                self.assertEqual(notification.target_user, user)
                self.assertEqual(notification.status, Notification.Status.CREATED)

                # Check channel and template are correct
                channel = Channel(notification)
                self.assertIsInstance(channel.message, TaskTestMessage)

            self.assertEqual(len(users), len(notifications))

            # Trigger task to process all remaining notifications
            processed_count = tasks.process_all_notifications()
            self.assertEqual(processed_count, len(notifications))

            # Check status
            for notification in notifications:
                notification = Notification.objects.get(pk=notification.id)
                self.assertEqual(notification.status, Notification.Status.SUCCESSFULLY_SENT)
                self.assertGreaterEqual(notification.datetime_processed, notification.datetime_scheduled)

            # Check email was sent
            self.assertEqual(len(mail.outbox), len(notifications))
            self.assertEqual(mail.outbox[0].subject, TRIGGER_SUBJECT)


            # Trigger task ... again
            processed_count = tasks.process_all_notifications()
            self.assertEqual(processed_count, 0)
            self.assertEqual(len(mail.outbox), len(notifications))


    def test_process_all_notifications_task_future_one(self):

        with self.settings(TRANSMISSIONS_TRIGGERS=self.trigger_settings):
            # Create trigger
            trigger_name = TRIGGER_NAME

            # Schedule notification for later
            user = factories.User()
            later_notification = TaskTestMessage.trigger(user, datetime_scheduled=timezone.now()+timezone.timedelta(days=2))
            self.assertEqual(later_notification.trigger_name, trigger_name)
            self.assertEqual(later_notification.target_user, user)
            self.assertEqual(later_notification.status, Notification.Status.CREATED)

            # Trigger task to process all remaining notifications
            processed_count = tasks.process_all_notifications()
            self.assertEqual(processed_count, 0)

            # Check status for later notification
            self.assertEqual(later_notification.status, Notification.Status.CREATED)
            self.assertIsNone(later_notification.datetime_processed)

            # Check no email was sent
            self.assertEqual(len(mail.outbox), 0)



    def test_process_all_notifications_task_except_future_one(self):

        with self.settings(TRANSMISSIONS_TRIGGERS=self.trigger_settings):
            users = [factories.User() for i in xrange(3)]
            notifications = []

            # Create trigger
            trigger_name = TRIGGER_NAME

            # Trigger notification
            for user in users:
                notification = TaskTestMessage.trigger(user)
                notifications.append(notification)
                self.assertEqual(notification.trigger_name, trigger_name)
                self.assertEqual(notification.target_user, user)
                self.assertEqual(notification.status, Notification.Status.CREATED)

                # Check channel and template are correct
                channel = Channel(notification)
                self.assertIsInstance(channel.message, TaskTestMessage)

            self.assertEqual(len(users), len(notifications))

            # Schedule notification for later
            user = factories.User()
            later_notification = TaskTestMessage.trigger(user, datetime_scheduled=timezone.now()+timezone.timedelta(days=2))
            self.assertEqual(later_notification.trigger_name, trigger_name)
            self.assertEqual(later_notification.target_user, user)
            self.assertEqual(later_notification.status, Notification.Status.CREATED)

            # Trigger task to process all remaining notifications
            processed_count = tasks.process_all_notifications()
            self.assertEqual(processed_count, len(notifications))

            # Check status
            for notification in notifications:
                notification = Notification.objects.get(pk=notification.id)
                self.assertEqual(notification.status, Notification.Status.SUCCESSFULLY_SENT)
                self.assertGreaterEqual(notification.datetime_processed, notification.datetime_scheduled)

            # Check email was sent
            self.assertEqual(len(mail.outbox), len(notifications))
            self.assertEqual(mail.outbox[0].subject, TRIGGER_SUBJECT)

            # Check status for later notification
            self.assertEqual(later_notification.status, Notification.Status.CREATED)
            self.assertIsNone(later_notification.datetime_processed)

            # Trigger task to process all remaining notifications
            processed_count = tasks.process_all_notifications()
            self.assertEqual(processed_count, 0)

    def test_process_broken(self):

        user = factories.User()

        # Trigger notification
        notification = BrokenMessage.trigger(user)
        self.assertEqual(notification.trigger_name, 'broken')
        self.assertEqual(notification.target_user, user)
        self.assertEqual(notification.status, Notification.Status.CREATED)

        # Trigger single task
        with self.assertRaisesMessage(Exception, 'This message is broken at init'):
            tasks.process_all_notifications()
        notification = Notification.objects.get(pk=notification.id)

        # Check that the notification is not retried
        num = tasks.process_all_notifications()
        self.assertEqual(num, 0)

        # Check status
        self.assertEqual(notification.status, Notification.Status.BROKEN)
        self.assertGreaterEqual(notification.datetime_processed, notification.datetime_scheduled)

        # Check email was sent
        self.assertEqual(len(mail.outbox), 0)


import logging

from django.test import TestCase
from django.core import mail

from transmissions import message
from transmissions.models import Notification
from transmissions.channels import Channel
from transmissions.channels.email import DefaultEmailMessage
from . import factories


TRIGGER_SIMPLE = 'simple-email'
@message(TRIGGER_SIMPLE, subject='Hello World!')
class SimpleMessage(DefaultEmailMessage):
    template_name = 'test'

class EmailChannelTests(TestCase):
    def setUp(self):
        logging.disable(logging.WARNING)

    def test_email_channel(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):
            user = factories.User()

            # Trigger notification
            notification = SimpleMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, SimpleMessage)


    def test_template_class(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            # Trigger notification
            notification = SimpleMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)
            self.assertIsNone(notification.content)
            diff = notification.datetime_created - notification.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, DefaultEmailMessage)

    def test_send_plain_email(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            # Trigger notification
            notification = SimpleMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)
            self.assertIsNone(notification.content)
            diff = notification.datetime_created - notification.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)
            self.assertEqual(notification.status, Notification.Status.CREATED)

            # Send notification
            notification.send()

            # Check status
            self.assertEqual(notification.status, Notification.Status.SUCCESSFULLY_SENT)
            self.assertGreaterEqual(notification.datetime_processed, notification.datetime_scheduled)

            # Check email was sent
            self.assertEqual(len(mail.outbox), 1)
            self.assertEqual(mail.outbox[0].subject, 'Hello World!')
            self.assertEqual(mail.outbox[0].body, '')
            self.assertFalse(hasattr(mail.outbox[0], 'alternatives'))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('transmissions', '0001_initial'),
    ]

    operations = [
        migrations.AlterIndexTogether(
            name='notification',
            index_together=set([('datetime_processed', 'datetime_scheduled'), ('target_user', 'datetime_scheduled'), ('target_user', 'trigger_name', 'datetime_processed')]),
        ),
        migrations.RemoveField(
            model_name='notification',
            name='trigger',
        ),
        migrations.DeleteModel(
            name='Trigger',
        ),
    ]



# -*- coding: utf-8 -*-
"""
    django-transmissions.models
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This is the core of django-transmissions with the following key models:
       * `Notification`: single communication events to a user with a `Trigger` on a given `Channel`
       * `Trigger`: Rule-set to create a `Notification`


    Additionally, a Channel is a mechanism or platform to send a Notification.
    Most commonly email or mobile push for iOS or Android.
"""

import pickle
from base64 import b64decode, b64encode

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.db import models
from django.utils import timezone

from django_enumfield import enum
from django_extensions.db import fields
from transmissions.channels import Channel
from transmissions.exceptions import ChannelSendException

if hasattr(settings, 'TRANSMISSION_USER_MODEL'):
    USER_MODEL = settings.TRANSMISSION_USER_MODEL
else:
    USER_MODEL = settings.AUTH_USER_MODEL


class BaseModel(models.Model):

    class Meta:
        abstract = True

    def get_admin_url(self):
        def view_name_for_model(model):
            return "admin:%s_%s_change" % (model._meta.app_label, model._meta.model_name)

        return reverse(view_name_for_model(self), args=(self.id,))


class Notification(BaseModel):

    """
    The instance of a message triggered to a user

    Notifications are single communication events so that we can track:
      * when it was triggered (by whom)
      * when was it scheduled for
      * when was it processed

    `Notification` are submitting a `Trigger` through a `Channel`
    """
    class Status(enum.Enum):
        CREATED = 0
        FAILED = -1
        CANCELLED = -2
        BROKEN = -3
        SUCCESSFULLY_SENT = 1

    uuid = fields.ShortUUIDField(unique=True, editable=False)

    trigger_name = models.CharField(db_index=True, max_length=50)

    target_user = models.ForeignKey(USER_MODEL, related_name='notifications')
    trigger_user = models.ForeignKey(USER_MODEL, related_name='notifications_sent',
                                     null=True, default=None)

    content_type = models.ForeignKey(ContentType, null=True, blank=True)
    content_id = models.PositiveIntegerField(null=True, blank=True)
    content = GenericForeignKey('content_type', 'content_id')
    data_pickled = models.TextField(blank=True, editable=False)

    datetime_created = models.DateTimeField(null=True, auto_now_add=True)
    datetime_scheduled = models.DateTimeField(db_index=True)
    datetime_processed = models.DateTimeField(db_index=True, null=True)
    datetime_seen = models.DateTimeField(null=True)
    datetime_consumed = models.DateTimeField(null=True)

    status = enum.EnumField(Status, default=Status.CREATED)

    @property
    def data(self):
        if not hasattr(self, '_data'):
            if len(self.data_pickled) <= 0:
                self._data = {}
            else:
                self._data = pickle.loads(b64decode(self.data_pickled.encode()))
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    class Meta:
        index_together = [['datetime_processed', 'datetime_scheduled'],
                          ['target_user', 'datetime_scheduled'],
                          ['target_user', 'trigger_name', 'datetime_processed']]
        app_label = 'transmissions'

    def send(self):
        """ Process notification and send via designated channel
        """

        try:
            channel = Channel(self)
            # Notification is not needed anymore
            if not channel.check_validity():
                self.status = self.Status.CANCELLED
            else:
                channel.send()
                self.status = self.Status.SUCCESSFULLY_SENT
            if channel.message.behavior == TriggerBehavior.DELETE_AFTER_PROCESSING:
                self.delete()
        except ChannelSendException:
            self.status = self.Status.FAILED
        except:
            self.status = self.Status.BROKEN
            raise
        finally:
            if self.pk:
                self.datetime_processed = timezone.now()
                self.save()

    def cancel(self):
        self.datetime_processed = timezone.now()
        self.status = self.Status.CANCELLED
        self.save()

    def save(self, *args, **kwargs):
        """
        Store pickled data before saving
        """

        self.data_pickled = b64encode(pickle.dumps(self.data)).decode()
        super(Notification, self).save(*args, **kwargs)

    def __unicode__(self):
        return u'Notification #{} to user #{}: {}'.format(
            self.pk, self.target_user_id, self.trigger_name)


class TriggerBehavior(enum.Enum):
    """
    Unless otherwise specified, Trigger Behaviors look for the existence of
    notifications with the same triger name and same addressee.
    """
    # Delete the Notification after it's processed.
    DELETE_AFTER_PROCESSING = 0
    DEFAULT = 10
    # There can be only one unprocessed notification at a time.
    TRIGGER_ONCE = 20
    # There can be only one unprocessed notification for this content at a
    # time.
    TRIGGER_ONCE_PER_CONTENT = 25
    # There can be only one notification ever sent.
    SEND_ONCE = 30
    # There can be only one notification for this content ever sent.
    SEND_ONCE_PER_CONTENT = 35
    # Cancel all pending notifications when a new one is scheduled.
    LAST_ONLY = 40



# -*- coding: utf-8 -*-
"""
    django-transmissions.channels
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Base class for channels

"""
import logging

from django.utils.module_loading import import_string
from django_enumfield import enum
from transmissions.exceptions import UnknownTriggerException, ChannelSendException


class Channel(object):

    def __init__(self, notification):
        self.notification = notification

        template_class = self.get_template()
        self.message = template_class(self.notification)

    class Types(enum.Enum):
        EMAIL = 1
        SMS = 2

    def get_template(self):

        # Dynamically load template class
        from transmissions.trigger import register
        if self.notification.trigger_name in register:
            return import_string(register[self.notification.trigger_name])

        raise UnknownTriggerException()

    def check_validity(self):
        """ Method called just before send() to determine if notification should still be sent

        :return: True if still valid, False if notification should be cancelled
        """
        return not hasattr(self.message, 'check_validity') or self.message.check_validity()

    def send(self):
        """ Send notification

        :param notification:
        :return:
        """

        try:
            return self.message.send()
        except Exception as e:
            logging.getLogger('django-transmissions').exception(e)
            raise ChannelSendException()



# -*- coding: utf-8 -*-
"""
    django-transmissions.tasks
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tasks to run asynchronously via Celery
"""

from transmissions.lock import lock
from django.utils import timezone
from celery.task import task

@task(ignore_result=True)
def process_notification(notification_id):
    from transmissions.models import Notification
    with lock('{0}'.format(notification_id)):
        # Load notification
        notification = Notification.objects.get(pk=notification_id)

        # Process if not processed already
        if notification.status == Notification.Status.CREATED:
            notification.send()


@task(ignore_result=True, time_limit=55)
def process_all_notifications():
    from transmissions.models import Notification
    notification_ids = Notification.objects.filter(datetime_scheduled__lte=timezone.now(),
                                                   datetime_processed__isnull=True).order_by('datetime_scheduled')\
        .values_list('id', flat=True)

    for notification_id in notification_ids:
        process_notification.delay(notification_id)

    return len(notification_ids)



import logging

from django.test import TestCase

from transmissions import message
from transmissions.channels.sms import DefaultSMSMessage
from transmissions.channels import Channel
from . import factories


TRIGGER_SIMPLE = 'simple-sms'
@message(TRIGGER_SIMPLE, subject='Hello World!')
class SimpleMessage(DefaultSMSMessage):
    template_name = 'test'

class SmsChannelTests(TestCase):
    def setUp(self):
        logging.disable(logging.WARNING)

    def test_sms_channel(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):
            user = factories.User()

            # Trigger notification
            notification = SimpleMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, DefaultSMSMessage)

    def test_template_class(self):

        welcome_path = "{}.{}".format(SimpleMessage.__module__, SimpleMessage.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(TRANSMISSIONS_TRIGGERS=trigger_settings):

            user = factories.User()

            # Trigger notification
            notification = SimpleMessage.trigger(user)
            self.assertEqual(notification.trigger_name, TRIGGER_SIMPLE)
            self.assertEqual(notification.target_user, user)
            self.assertIsNone(notification.content)
            diff = notification.datetime_created - notification.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)

            # Check channel and template are correct
            channel = Channel(notification)
            self.assertIsInstance(channel.message, DefaultSMSMessage)
